import os

import logomaker
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
from scipy import stats
import seaborn as sns
from sklearn.metrics import confusion_matrix

from cleavenet import analysis


def plot_parity(labels, predicted, mmps, save_path, nrows=6, ncols=3):
    """ Plot scatterplot of true vs. predicted z-scores and save .png
        Args:
        labels (np.array): true z-scores
        predicted (np.array): predicted z-scores
        mmp (list) : list of all mmps
        save_path (str): saving path
    """
    for i in range(len(mmps)):
        data = pd.DataFrame({'True': labels[:,i], 'Predicted': predicted[:,i]})
        p = sns.jointplot(data=data, x='True', y='Predicted', kind='reg') #, xlim=(-2, 5), ylim=(-2, 5))
        #corr_pear = data.corr(method='pearson')
        r_squared = stats.pearsonr(labels[:,i], predicted[:,i]).statistic
        label = " $R$=%.2f" % (r_squared)
        #ax.text(0.05, 0.95, label, transform=ax.transAxes, fontsize=14,
        #        verticalalignment='top')
        p.fig.suptitle(mmps[i]+label, weight='bold') #+' R\N{SUPERSCRIPT TWO}=%.2f' % corr_pear.iloc[0, 1])
        p.fig.tight_layout()
        save_dir_test = os.path.join(save_path, mmps[i]+'_true_pred_scatter.svg')
        p.savefig(save_dir_test)


def plot_rmse(rmse, mmps, save_path):
    "Plot average RMSE score for each MMP family after training"
    rmse = np.array(rmse)
    fig = plt.figure(figsize=(12,3))
    # creating the bar plot
    plt.bar(mmps, rmse, color='k', alpha=0.75, width=0.75)
    #plt.xlabel("MMP Family")
    plt.ylabel("RMSE")
    plt.ylim(0,1)
    plt.tight_layout()
    np.savetxt(save_path + 'rmse.csv', rmse, delimiter=", ", fmt='% s')
    save_file = save_path + 'rmse.svg'
    fig.savefig(save_file)


def plot_mae(mae, mmps, save_path):
    "Plot average MAE score for each MMP family after training"
    mae = np.array(mae)
    fig = plt.figure(figsize=(12,3))
    plt.bar(mmps, mae, color='k', alpha=0.75, width=0.75)
    #plt.xlabel("MMP Family")
    plt.ylabel('MAE')
    plt.ylim(0,1)
    plt.tight_layout()
    np.savetxt(save_path+'mae.csv', mae, delimiter=", ", fmt='% s')
    save_file = save_path+'mae.svg'
    fig.savefig(save_file)


def true_pred_ranked_scatter_z(scores, save_path, mmp):
    "Plot ranked predicted z-scores colored by true z-score"
    fig, ax = plt.subplots(figsize=(5, 5))
    colors = {0: 'blue', 1: 'red'}
    scores_sorted = scores.sort_values(by=['Pred Z-scores'], ascending=True)
    ax.scatter(np.arange(len(scores_sorted['Pred Z-scores'].values)),
               scores_sorted['Pred Z-scores'].values,
               c=scores_sorted['Cleaved true'].map(colors))
    plt.axhline(y=0, c='k', alpha=0.50, ls='dashed')
    plt.xlabel('Ranked substrates by predicted z-score')
    plt.ylabel('Pred z-score')
    plt.title(mmp, weight='bold')
    save_file='true_pred_ranked_z.svg'
    save_dir_test = os.path.join(save_path, mmp+save_file)
    fig.savefig(save_dir_test)
    plt.close()


def confidence_ranked_scatter_z(scores, save_path, mmp, threshold=0):
    "Plot ranked z-scores colored by confidence score"
    fig, ax = plt.subplots(figsize=(6, 5))
    vmin = 0
    vmax = 0.5 # scores['uncertainty'].max()
    ymin = scores['Pred Z-scores'].min()-0.25
    ymax = scores['Pred Z-scores'].max()+0.25
    if threshold > 0:
        scores = scores[scores['uncertainty']<=threshold]
        save_file = 'confidence_ranked_z_threshold.svg'
    else:
        save_file = 'confidence_ranked_z.svg'
    scores_sorted = scores.sort_values(by=['Pred Z-scores'], ascending=True)
    p = ax.scatter(np.arange(len(scores_sorted['Pred Z-scores'].values)),
               scores_sorted['Pred Z-scores'].values,
               c=scores_sorted['uncertainty'], vmin=vmin, vmax=vmax)
    plt.axhline(y=0, c='k', alpha=0.50, ls='dashed')
    plt.ylim(ymin, ymax)
    fig.colorbar(p, ax=ax)
    plt.xlabel('Ranked substrates by predicted z-score')
    plt.ylabel('Pred z-score')
    plt.title(mmp, weight='bold')
    save_dir_test = os.path.join(save_path, mmp+save_file)
    fig.savefig(save_dir_test)
    plt.close()


def confusion(y_true, y_pred, save_path, mmp):
    """ Calculate, plot and save confusion matrix

        Args:
        y_true (np.array): true labels
        y_pred (float): predicted labels

        Saves:
        confusion_matrix.png
        confusion_matrix.csv
    """
    # Calculate confusion matrix
    labels = ['Non-cleaved', 'Cleaved']
    cm = confusion_matrix(y_true, y_pred, normalize='true')

    # Save heatmap plot of confusion matrix
    fig = plt.figure()
    ax = plt.subplot()
    sns.heatmap(cm, cmap=sns.color_palette("Blues", as_cmap=True),annot=True, ax=ax)  # annot=True to annotate cells
    # labels, title and ticks
    ax.set_xlabel('Predicted labels')
    ax.set_ylabel('True labels')
    ax.set_title(mmp, weight='bold')
    ax.xaxis.set_ticklabels(labels, size=13)
    ax.yaxis.set_ticklabels(labels, size=13)
    fig.savefig(os.path.join(save_path, mmp+'_confusion_matrix.svg'))
    plt.close()


def confidence_histogram(scores, save_path, mmp):
    ci = analysis.confidence_interval_1_sided(np.std(scores['uncertainty']), len(scores), confidence_level=0.80)
    #print(scores['uncertainty'].mean()+ci)
    fig, ax = plt.subplots(figsize=(6, 5))
    sns.histplot(data=scores, x='uncertainty', ax=ax, kde=True)
    plt.axvline(x=scores['uncertainty'].mean()+ci, c='k', alpha=0.50, ls='dashed')
    plt.xlabel('Uncertainty')
    plt.ylabel('Count')
    plt.title(mmp)
    save_dir_test = os.path.join(save_path, mmp + '_confidence_hist.svg')
    fig.savefig(save_dir_test)
    plt.close()
    return(scores['uncertainty'].mean()+ci)


def confidence_test_histogram(scores, save_path, mmp):
    fig, ax = plt.subplots(figsize=(6, 5))
    sns.regplot(data=scores, x='True Z-scores', y='Pred Z-scores', ax=ax, ci=0.8)
    plt.xlabel('Uncertainty')
    plt.ylabel('Count')
    plt.title(mmp)
    save_dir_test = os.path.join(save_path, mmp + '_confidence_hist.svg')
    fig.savefig(save_dir_test)
    plt.close()


def heatmaps(save_path, data_dir = '/generated/', cbar_center=0, weighted=False, cluster=True, top=50):
    plot_mmps = ['MMP1', 'MMP2', 'MMP3', 'MMP8', 'MMP9', 'MMP10', 'MMP11',
            'MMP12', 'MMP13', 'MMP14', 'MMP15', 'MMP16', 'MMP17', 'MMP20',
            'MMP24', 'MMP25', 'MMP7'] #, 'MMP19'] # Exclude MMP19
    if weighted:
        all_scores = pd.read_csv(data_dir+'weighted_all_scores.csv')
    else:
        all_scores = pd.read_csv(data_dir+'all_scores.csv')
    dfs = []
    if weighted:
        file_name = '_weighted_top' + str(top) + '_cleaved.csv'
        if cluster:
            save_name = 'weighted_cluster_heatmap.svg'
            save_df = 'weighted_cluster_df.csv'
            save_col = 'weighted_cluster_col.csv'
            save_row = 'weighted_cluster_row.csv'
        else:
            save_df = 'weighted_df.csv'
            save_name = 'weighted_heatmap.svg'
            save_col = 'weighted_col.csv'
            save_row = 'weighted_row.csv'
    else:
        file_name = '_top' + str(top) + '_cleaved.csv'
        if cluster:
            save_name = 'top_cluster_heatmap.svg'
            save_df = 'top_cluster_df.csv'
            save_col = 'top_cluster_col.csv'
            save_row = 'top_cluster_row.csv'
        else:
            save_df = 'top_df.csv'
            save_name = 'top_heatmap.svg'
            save_col = 'top_col.csv'
            save_row = 'top_row.csv'
    for mmp in plot_mmps:
        file_path = data_dir + mmp + file_name
        df_temp = pd.read_csv(file_path)[:top] # top 20 for each MMP
        dfs.append(df_temp)
    df = pd.concat(dfs)
    df_merge = df.merge(all_scores, how='left')
    df_merge = df_merge.drop_duplicates(subset=['sequences']) # drop duplicate sequences, add additional scores
    fixed_seq_list = [seq[1:] for seq in list(df_merge['sequences'])]
    df_merge['sequences'] = fixed_seq_list
    df_merge = df_merge.set_index('sequences')
    print(df_merge)
    # Default plot
    if cluster:
        ax = sns.clustermap(df_merge[plot_mmps], cmap='RdBu_r', center=cbar_center, figsize=(12,10)) #, vmin=-2, vmax=5)
        cluster_row = ax.dendrogram_row.reordered_ind
        cluster_col = ax.dendrogram_col.reordered_ind
        df_merge.to_csv(save_path+save_df)
        np.savetxt(save_path+save_row, cluster_row, delimiter=", ", fmt='% s')
        np.savetxt(save_path+save_col, cluster_col, delimiter=", ", fmt='% s')
    else:
        fig = plt.figure()
        ax = plt.subplot()
        sns.heatmap(df_merge[plot_mmps], cmap='RdBu_r', center=cbar_center) #, vmin=-2, vmax=5)
        plt.tight_layout()
    save_plot = os.path.join(save_path, save_name)
    plt.savefig(save_plot)


def aa_freq(x_train, x_gen, dataloader, save_path, c='xkcd:pea', save_file='aa_frequency.svg'):
    train_seqs = analysis.extract_sequences(x_train, dataloader)
    if x_gen is not None:
        gen_seqs = analysis.extract_sequences(x_gen, dataloader)
        gen_dict_arr = analysis.count_by_res(gen_seqs)
    dict_arr = analysis.count_by_res(train_seqs)
    # Plot
    nrow = 2
    ncol = 5
    title = ['P5', 'P4', 'P3', 'P2', 'P1', "P1'", "P2'", "P3'", "P4'", "P5'"]
    fig, axs = plt.subplots(nrow, ncol, figsize=(20, 5.5), sharey=False, sharex=False)

    all_kl = []
    all_norm = []
    for i, ax in enumerate(fig.axes):
        if (i >= 2 and i <= 6):
            border_color = 'firebrick'
        else:
            border_color = 'k'
        ax.spines['bottom'].set_color(border_color)
        ax.spines['top'].set_color(border_color)
        ax.spines['left'].set_color(border_color)
        ax.spines['right'].set_color(border_color)
        plt.setp(ax.spines.values(), linewidth=2.5)

        ax.set_title(title[i], weight='bold', fontsize=18)
        ax.set_yticks([])
        keys = list(dict_arr[i].keys())
        values = list(dict_arr[i].values())
        norm = analysis.normalize(values)
        all_norm.append(norm)
        ax.bar(keys, norm, color='xkcd:mid blue', alpha=0.75)
        
        if x_gen is not None:
            if '$' in list(gen_dict_arr[i].keys()):
                gen_dict_arr[i].pop('$')
            g_keys = list(gen_dict_arr[i].keys())
            g_values = list(gen_dict_arr[i].values())
            g_norm = analysis.normalize(g_values)
            ax.bar(g_keys, g_norm, color=c, alpha=0.75)

            kl = scipy.special.kl_div(norm, g_norm)
            kl = [k for k in kl if k !=float('inf')]
            kl = sum(kl)
            all_kl.append(kl)
            kl_label = "$KL$=%.3f" % (kl)
            ax.text(0.65, 0.95, kl_label, transform=ax.transAxes, fontsize=14,
                verticalalignment='top')

            for label in (ax.get_xticklabels() + ax.get_yticklabels()): label.set_fontsize(16)

    if x_gen is not None:
        print("KL:", np.mean(all_kl))
    fig.axes[0].set_ylabel('Frequency', fontsize=18)
    fig.axes[5].set_ylabel('Frequency', fontsize=18)
    fig.tight_layout()
    save_plot = os.path.join(save_path, save_file)
    plt.savefig(save_plot)
    if x_gen is None:
        char2idx = {u: i for i, u in enumerate(keys)}
        idx2char = np.array(list(keys))
        return all_norm, char2idx, idx2char


def z_score_distribution(y_train, y_pred, mmps, save_path):
    fig, axes = plt.subplots(nrows=5, ncols=4, figsize=(18, 20))
    for i, ax in enumerate(axes.ravel()[:18]):
        ax2 = ax.twinx()
        ax.hist(y_train[:, i], bins=20, color='xkcd:mid blue', label='train', alpha=0.75)
        ax2.hist(y_pred[:,i], bins=20, color='xkcd:pea', label='generated', alpha=0.75)
        ax.set_xlabel('Z-score')
        ax.set_ylabel('Count')
        ax.set_title(mmps[i], weight='bold', fontsize=16)
        ax2.axis('off')
        ax.set_yticks([])
    fig.delaxes(axes[4][2])
    fig.delaxes(axes[4][3])
    fig.tight_layout()
    save_plot = os.path.join(save_path, 'z_score_distribution.svg')
    plt.savefig(save_plot)


def plot_auc(y, y_hat, mmp_index, mmps, save_path):
    z_score_list = [0.0, 1.0, 1.5, 2.0, 2.5]
    alphas = [1, 0.8, 0.6, 0.4, 0.2]
    fig, ax = plt.subplots()
    for i, z_score_cutoff in enumerate(z_score_list):
        fpr, tpr = analysis.get_roc_data(y, y_hat, true_z_cutoff=z_score_cutoff)
        auc = analysis.calc_auc(fpr, tpr)
        label = " Z-score:{}, AUC:{:2.2}".format(z_score_cutoff, auc)

        ax.plot(fpr, tpr, label=label, c='#124b9a', alpha=alphas[i])
        ax.set_title(mmps[mmp_index])
        lims = [
            np.min([ax.get_xlim(), ax.get_ylim()]),  # min of both axes
            np.max([ax.get_xlim(), ax.get_ylim()]),  # max of both axes
        ]
        ax.set_aspect('equal')
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)
        ax.set_ylabel('True positive rate')
        ax.set_xlabel('False positive rate')
    ax.plot(lims, lims, 'k--', alpha=0.75, zorder=0)
    ax.legend(loc='lower right')
    save_plot = os.path.join(save_path, mmps[mmp_index]+'_auc.svg')
    plt.savefig(save_plot)


def aa_frequencies(zscore):
    """
    Calculate amino acid frequencies at each position in a set of aligned sequences.

    Parameters:
    - zscore (pd.DataFrame): DataFrame containing a column 'sequence' with aligned amino acid sequences.

    Returns:
    - pd.DataFrame: A DataFrame with amino acids as rows and sequence positions as columns,
                    representing the frequency of each amino acid at each position.
    """
    # Step 1: Initialize parameters
    sequence_length = len(zscore['sequence'].iloc[0])  # Assuming all sequences are the same length
    amino_acids = 'ACDEFGHIKLMNPQRSTVWY'

    # Step 2: Create a frequency matrix initialized to zero
    frequency_matrix = pd.DataFrame(0, index=list(amino_acids), columns=range(1, sequence_length + 1))

    # Step 3: Populate the frequency matrix
    for sequence in zscore['sequence']:
        for position, amino_acid in enumerate(sequence, start=1):
            if amino_acid in amino_acids:  # Only process valid amino acids
                frequency_matrix.loc[amino_acid, position] += 1
            else:
                print(f"Warning: Unexpected character '{amino_acid}' in sequence '{sequence}' at position {position}")

    # Step 4: Normalize the frequencies by the total number of sequences
    frequency_matrix = frequency_matrix.div(len(zscore))

    # Step 5: Reorder the rows based on amino acid properties
    reordered_amino_acids = [
        'F', 'W', 'Y',  # Hydrophobic Aromatic
        'P', 'A', 'I', 'L', 'M', 'V', 'G',  # Hydrophobic
        'C', 'N', 'Q', 'S', 'T',  # Hydrophilic
        'D', 'E', 'H',  # Acidic
        'K', 'R'  # Basic
    ]
    frequency_matrix = frequency_matrix.reindex(reordered_amino_acids)

    # Step 6: Rename the columns to reflect sequence positions
    frequency_matrix.columns = ['P5', 'P4', 'P3', 'P2', 'P1', "P1'", "P2'", "P3'", "P4'", "P5'"]

    return frequency_matrix


def aa_frequencies(zscore):
    """
    Calculate amino acid frequencies at each position in a set of aligned sequences.

    Parameters:
    - zscore (pd.DataFrame): DataFrame containing a column 'sequence' with aligned amino acid sequences.

    Returns:
    - pd.DataFrame: A DataFrame with amino acids as rows and sequence positions as columns,
                    representing the frequency of each amino acid at each position.
    """
    # Step 1: Initialize parameters
    sequence_length = len(zscore['sequence'].iloc[0])  # Assuming all sequences are the same length
    amino_acids = 'ACDEFGHIKLMNPQRSTVWY'

    # Step 2: Create a frequency matrix initialized to zero
    frequency_matrix = pd.DataFrame(0, index=list(amino_acids), columns=range(1, sequence_length + 1))

    # Step 3: Populate the frequency matrix
    for sequence in zscore['sequence']:
        for position, amino_acid in enumerate(sequence, start=1):
            if amino_acid in amino_acids:  # Only process valid amino acids
                frequency_matrix.loc[amino_acid, position] += 1
            else:
                print(f"Warning: Unexpected character '{amino_acid}' in sequence '{sequence}' at position {position}")

    # Step 4: Normalize the frequencies by the total number of sequences
    frequency_matrix = frequency_matrix.div(len(zscore))

    # Step 5: Reorder the rows based on amino acid properties
    reordered_amino_acids = [
        'F', 'W', 'Y',  # Hydrophobic Aromatic
        'P', 'A', 'I', 'L', 'M', 'V', 'G',  # Hydrophobic
        'C', 'N', 'Q', 'S', 'T',  # Hydrophilic
        'D', 'E', 'H',  # Acidic
        'K', 'R'  # Basic
    ]
    frequency_matrix = frequency_matrix.reindex(reordered_amino_acids)

    # Step 6: Rename the columns to reflect sequence positions
    frequency_matrix.columns = ['P5', 'P4', 'P3', 'P2', 'P1', "P1'", "P2'", "P3'", "P4'", "P5'"]

    return frequency_matrix


def normalize_matrix(frequency_matrix, reference_df):
    """
    Normalize the frequency matrix using a reference DataFrame.

    Parameters:
    - frequency_matrix (pd.DataFrame): DataFrame where rows represent amino acids and columns represent sequence positions, containing frequencies.
    - reference_df (pd.DataFrame): DataFrame with the same shape as frequency_matrix, representing the reference background frequencies.

    Returns:
    - pd.DataFrame: Information content matrix after normalization.
    """

    def calculate_ic(frequency: float, background: float) -> float:
        """Calculate information content given frequency and background frequency."""
        return frequency * np.log2(frequency / background) if frequency > 0 else 0

    assert frequency_matrix.shape == reference_df.shape
    assert (frequency_matrix.index == reference_df.index).all()
    assert (frequency_matrix.columns == reference_df.columns).all()

    ic_matrix_normalized = frequency_matrix.copy()  # Create a copy to avoid modifying the original DataFrame
    for i in frequency_matrix.index:
        for j in frequency_matrix.columns:
            ic_matrix_normalized.at[i, j] = calculate_ic(frequency_matrix.at[i, j], reference_df.at[i, j])

    return ic_matrix_normalized


def plot_icelogo(frequency_matrix, save_path, custom_color_scheme, source=None, name=None):
    """
    Generate and save different types of sequence logos based on frequency and information content.

    Parameters:
    - frequency_matrix (pd.DataFrame): DataFrame where rows represent amino acids and columns represent sequence positions, containing frequencies.
    - bg_df (pd.DataFrame): DataFrame with the same shape as frequency_matrix, representing background frequencies in training.
    - nat_df (pd.DataFrame): DataFrame with the same shape as frequency_matrix, representing background frequencies in natural protein sequences.
    - save_path (str): Directory path where the logo images will be saved.
    - custom_color_scheme (dict): Dictionary containing color for each amino acid.
    - source (str, optional): Name from data source
    - name (str, optional): Name used for the logo files and titles.

    Returns:
    - None: The function saves the sequence logos to the specified path.
    """

    # 1. Generate and save un-normalized sequence logo
    freq_logo = logomaker.Logo(frequency_matrix,
                               shade_below=.5,
                               fade_below=.5,
                               font_name='Arial Rounded MT Bold',
                               color_scheme=custom_color_scheme, figsize=(6,2))
    freq_logo.style_spines(visible=False)
    freq_logo.style_spines(spines=['left', 'bottom'], visible=True)
    freq_logo.style_xticks(rotation=90, fmt='%d', anchor=0)
    freq_logo.ax.set_ylabel(f"Frequency - {name}", labelpad=-1)
    freq_logo.ax.set_xticklabels(['P5', 'P4', 'P3', 'P2', 'P1', "P1'", "P2'", "P3'", "P4'", "P5'"])
    plt.tight_layout()
    plt.savefig(os.path.join(save_path, f"{source}_logo_{name}.png"), dpi=300)


import os

import numpy as np
import tensorflow as tf
from tensorflow.keras.regularizers import L2

import cleavenet.data
from cleavenet import analysis, plotter
from cleavenet.utils import mmps 


class TransformerSchedule(tf.keras.optimizers.schedules.LearningRateSchedule):
    """
    Custom schedule from original Transformer paper https://arxiv.org/abs/1706.03762
    Adapted from https://www.tensorflow.org/text/tutorials/transformer#set_up_the_optimizer
    """
    def __init__(self, d_model, warmup_steps=4000):
        super().__init__()
        self.d_model = d_model
        self.d_model = tf.cast(self.d_model, tf.float32)
        self.warmup_steps = warmup_steps

    def __call__(self, step):
        step = tf.cast(step, dtype=tf.float32)
        arg1 = tf.math.rsqrt(step)
        arg2 = step * (self.warmup_steps ** -1.5)
        #print(tf.math.rsqrt(self.d_model) * tf.math.minimum(arg1, arg2))
        return tf.math.rsqrt(self.d_model) * tf.math.minimum(arg1, arg2)


def positional_encoding(length, depth):
    "positional encoding for transformer, from https://www.tensorflow.org/text/tutorials/transformer"
    depth = depth/2
    positions = np.arange(length)[:, np.newaxis]     # (seq, 1)
    depths = np.arange(depth)[np.newaxis, :]/depth   # (1, depth)
    angle_rates = 1 / (10000**depths)         # (1, depth)
    angle_rads = positions * angle_rates      # (pos, depth)
    pos_encoding = np.concatenate(
      [np.sin(angle_rads), np.cos(angle_rads)],
      axis=-1)
    return tf.cast(pos_encoding, dtype=tf.float32)


class PositionalEmbedding(tf.keras.layers.Layer):
    """
    Looks up a tokens embedding vector, and adds a positional encoding vector

    label=True when input has a conditional tag (z-scores)
    """
    def __init__(self, vocab_size, d_model, max_seq_length=15, mask_zero=True, label=False, start_idx=20):
        super().__init__()
        self.d_model = d_model
        self.embedding = tf.keras.layers.Embedding(vocab_size, d_model, mask_zero=mask_zero)
        self.pos_encoding = positional_encoding(length=max_seq_length, depth=d_model)
        self.label = label
        self.start_idx = start_idx
        if label: 
            self.label_embedding = tf.keras.layers.Dense(d_model)

    def call(self, x):
        if self.label:
            x, label = x 
            label_emb = []
            #print("label", label)
            for i in range(len(label)):
                if label[i][0] == self.start_idx: # if start token 
                    label_emb_temp = self.embedding(label[i])
                else:
                    label_emb_temp = self.label_embedding(tf.expand_dims(label[i], 0))
                #print("embedded", label_emb_temp)
                label_emb.append(label_emb_temp)
            label_emb = tf.stack(label_emb)
        x = self.embedding(x)
        if self.label:
            if x.shape[1] == 0:
                x = label_emb
            else:
                x = tf.concat([label_emb, x], 1)
        length = tf.shape(x)[1]
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        x = x + self.pos_encoding[tf.newaxis, :length, :]
        return x


### Transformer Components ###
class BaseAttention(tf.keras.layers.Layer):
    """
    Base attention layer, contains component layers;
    - contains a multihead attention, and layers.Add + layers.LayersNormalization
    e.g. Each Nx of a Transformer model
    """
    def __init__(self, **kwargs):
        super().__init__()
        self.mha = tf.keras.layers.MultiHeadAttention(**kwargs)
        self.layernorm = tf.keras.layers.LayerNormalization()
        self.add = tf.keras.layers.Add()


class GlobalSelfAttention(BaseAttention):
    """
    Global self attention layer; processing context sequence, and propegating information along length
    (information flows in both directions)

    x: target sequence (query, and in this case also the value)
    """
    def call(self, x):
        attn_output = self.mha(
            query=x,
            value=x,
            key=x)
        x = self.add([x, attn_output])
        x = self.layernorm(x)
        return x


class CausalSelfAttention(BaseAttention):
    """
    Causal self attention; similar to global self attention but for output sequence
    To build a casual self attention layer, use appropriate mask when computing attention scores and summing attention
    value
    """
    def call(self, x):
        attn_output, attn_scores = self.mha(
            query=x,
            value=x,
            key=x,
            use_causal_mask=True,
            return_attention_scores=True) # causal self attention layer; each location only has access to locations before it
        # Cache the attention scores for plotting later.
        self.last_attn_scores = attn_scores
        x = self.add([x, attn_output])
        x = self.layernorm(x)
        return x


class FeedForward(tf.keras.layers.Layer):
    """
    Feed forward network; consists of 2 linear layers, with RELU activation and dropout layer
    x: embedded input B, T, dim
    """
    def __init__(self, d_model, dff, activation='relu', dropout_rate=0.1):
        super().__init__()
        self.seq = tf.keras.Sequential([
          tf.keras.layers.Dense(dff, activation=activation),
          tf.keras.layers.Dense(d_model),
          tf.keras.layers.Dropout(dropout_rate)
        ])
        self.add = tf.keras.layers.Add()
        self.layer_norm = tf.keras.layers.LayerNormalization()

    def call(self, x):
        x = self.add([x, self.seq(x)])
        x = self.layer_norm(x)
        return x


class EncoderLayer(tf.keras.layers.Layer):
    """
    Encoder layer; Global self attention and feed forward
    """
    def __init__(self,*, d_model, num_heads, dff, activation='relu', dropout_rate=0.1):
        super().__init__()
        self.self_attention = GlobalSelfAttention(
            num_heads=num_heads,
            key_dim=d_model,
            dropout=dropout_rate)
        self.ffn = FeedForward(d_model, dff, activation=activation, dropout_rate=dropout_rate)

    def call(self, x):
        x = self.self_attention(x)
        x = self.ffn(x)
        return x


class Encoder(tf.keras.layers.Layer):
    """
    Encoder; Positional embedding, stack of N encoder layers
    """
    def __init__(self, *, num_layers, d_model, num_heads, dff, vocab_size, activation='relu', dropout_rate=0.1, mask_zero=False):
        super().__init__()
        self.d_model = d_model
        self.num_layers = num_layers
        self.pos_embedding = PositionalEmbedding(
            vocab_size=vocab_size, d_model=d_model, mask_zero=mask_zero)
        self.enc_layers = [
            EncoderLayer(d_model=d_model,
                         num_heads=num_heads,
                         dff=dff,
                         activation=activation,
                         dropout_rate=dropout_rate)
            for _ in range(num_layers)]
        self.dropout = tf.keras.layers.Dropout(dropout_rate)

    def call(self, x):
        # `x` is token-IDs shape: (batch, seq_len)
        x = self.pos_embedding(x)  # Shape `(batch_size, seq_len, d_model)`.
        x = self.dropout(x)
        for i in range(self.num_layers):
          x = self.enc_layers[i](x)
        return x  # Shape `(batch_size, seq_len, d_model)`.


class DecoderLayer(tf.keras.layers.Layer):
    "Decoder layer; Causal self attention, feed forward layer "
    def __init__(self, *, d_model, num_heads, dff, dropout_rate=0.1):
        super(DecoderLayer, self).__init__()
        self.causal_self_attention = CausalSelfAttention(
            num_heads=num_heads,
            key_dim=d_model,
            dropout=dropout_rate)
        self.ffn = FeedForward(d_model, dff)

    def call(self, x):
        x = self.causal_self_attention(x=x)
        self.last_attn_scores = self.causal_self_attention.last_attn_scores
        x = self.ffn(x)  # Shape `(batch_size, seq_len, d_model)`.
        return x


class Decoder(tf.keras.layers.Layer):
    """
    Decoder; positional embedding, stack of decoder layers
    """
    def __init__(self, *, num_layers, d_model, num_heads, dff, vocab_size, dropout_rate=0.1):
        super(Decoder, self).__init__()
        self.d_model = d_model
        self.num_layers = num_layers
        self.pos_embedding = PositionalEmbedding(vocab_size=vocab_size, d_model=d_model)
        self.dropout = tf.keras.layers.Dropout(dropout_rate)
        self.dec_layers = [
            DecoderLayer(d_model=d_model, num_heads=num_heads, dff=dff, dropout_rate=dropout_rate)
            for _ in range(num_layers)]
        self.last_attn_scores = None

    def call(self, x):
        # `x` is token-IDs shape (batch, target_seq_len)
        x = self.pos_embedding(x)  # (batch_size, target_seq_len, d_model)
        x = self.dropout(x)
        for i in range(self.num_layers):
          x  = self.dec_layers[i](x)
        self.last_attn_scores = self.dec_layers[-1].last_attn_scores
        return x #  (batch_size, target_seq_len, d_model).


### Transformer Models ###
class TransformerDecoder(tf.keras.Model):
    "Transformer; Decoder-only model for autoregressive modeling"
    def __init__(self, *, num_layers, d_model, num_heads, dff, vocab_size, dropout_rate=0.1):
        super().__init__()
        self.decoder = Decoder(num_layers=num_layers, d_model=d_model,
                               num_heads=num_heads, dff=dff,
                               vocab_size=vocab_size,
                               dropout_rate=dropout_rate)
        self.final_layer = tf.keras.layers.Dense(vocab_size)
        self.vocab_size=vocab_size

    def call(self, x):
        x = self.decoder(x) # (batch_size, target_len, d_model)
        logits = self.final_layer(x)  # (batch_size, target_len, target_vocab_size)
        return logits

    def compute_loss(self, y, y_hat):
        loss = tf.keras.losses.sparse_categorical_crossentropy(y, y_hat, from_logits=True)
        return tf.reduce_mean(loss)

    def compute_accuracy(self, y, y_hat):
        m = tf.keras.metrics.SparseCategoricalAccuracy()
        m.update_state(y, y_hat)
        return m.result()


class ConditionalTransformerDecoder(tf.keras.Model):
    "Transformer; Decoder-only model for autoregressive modeling"
    def __init__(self, *, num_layers, d_model, num_heads, dff, vocab_size, dropout_rate=0.1):
        super().__init__()
        self.d_model = d_model
        self.num_layers = num_layers
        self.pos_embedding = PositionalEmbedding(vocab_size=vocab_size, d_model=d_model, label=True)
        self.dropout = tf.keras.layers.Dropout(dropout_rate)
        self.dec_layers = [
            DecoderLayer(d_model=d_model, num_heads=num_heads, dff=dff, dropout_rate=dropout_rate)
            for _ in range(num_layers)]
        self.last_attn_scores = None
        self.vocab_size=vocab_size
        self.final_layer = tf.keras.layers.Dense(vocab_size)

    # def build(self, input_shape): # Don't use with new TF
    #     self.pos_embedding.build(input_shape)
    #     input_shape = self.pos_embedding.compute_output_shape(input_shape)
    #     self.dec_layers.build(input_shape)
    #     input_shape = self.dec_layers.compute_output_shape(input_shape)
    #     self.final_layer.build(input_shape)
    #     self.built = True

    def call(self, x):
        x = self.pos_embedding(x)  # (batch_size, target_seq_len, d_model)
        x = self.dropout(x)
        for i in range(self.num_layers):
          x  = self.dec_layers[i](x)
        self.last_attn_scores = self.dec_layers[-1].last_attn_scores
        logits = self.final_layer(x)  # (batch_size, target_len, target_vocab_size)
        return logits

    def compute_loss(self, y, y_hat):
        loss = tf.keras.losses.sparse_categorical_crossentropy(y, y_hat, from_logits=True)
        return tf.reduce_mean(loss)

    def compute_accuracy(self, y, y_hat):
        m = tf.keras.metrics.SparseCategoricalAccuracy()
        m.update_state(y, y_hat)
        return m.result()
        

class TransformerEncoder(tf.keras.Model):
    "Transformer; Encoder-only model for BERT MLM"
    def __init__(self, *, num_layers, d_model, num_heads, dff, vocab_size, dropout_rate=0.1, output_dim=None, pool_outputs=False, mask_zero=False):
        super().__init__()
        if output_dim == None:
            output_dim=vocab_size
        self.pool_outputs = pool_outputs
        self.encoder = Encoder(num_layers=num_layers, d_model=d_model,
                               num_heads=num_heads, dff=dff,
                               vocab_size=vocab_size,
                               dropout_rate=dropout_rate, mask_zero=mask_zero)
        self.final_layer = tf.keras.layers.Dense(output_dim)
        self.vocab_size = vocab_size

    def call(self, x):
        x = self.encoder(x)  # (batch_size, target_len, d_model)
        self.last_layer_embeddings = x
        if self.pool_outputs:
            x = tf.squeeze(x[:, 0:1, :], axis=1) # (batch_size, 1, output) , using for regression task (pool outputs by taking representation of first token)
        logits = self.final_layer(x)  # (batch_size, target_len, target_vocab_size)
        return logits

    def compute_masked_loss(self, y, y_hat, mask):
        """ Computes loss between true y and prediction over masked locations"""
        # Select masked locations
        y_masked = tf.boolean_mask(y, mask)
        y_hat_masked = tf.boolean_mask(y_hat, mask)
        loss = tf.keras.losses.sparse_categorical_crossentropy(y_masked, y_hat_masked, from_logits=True)
        return tf.reduce_mean(loss)

    def compute_masked_accuracy(self, y, y_hat, mask):
        m = tf.keras.metrics.CategoricalAccuracy()
        # Compute accuracy over masked tokens
        y_onehot = tf.one_hot(y, depth=self.vocab_size)
        y_masked = tf.boolean_mask(y_onehot, mask)
        y_hat_masked = tf.boolean_mask(y_hat, mask)

        # Compute accuracy
        m.update_state(y_masked, y_hat_masked)
        accuracy = m.result()
        return accuracy

    def compute_loss(self, y, y_hat):
        """ Computes loss, as mean absolute error, between true y and prediction.

        Args:
            y: true labels
            y_hat: output prediction

        Returns:
            absolute error
        """
        y_hat = y_hat # batch, last, 1
        y_hat = tf.cast(y_hat, tf.float64)
        error = tf.abs(y - y_hat)
        return tf.reduce_mean(error)

    def compute_rmse(self, y, y_hat, axis=None):
        """Computes the root mean squared error between true y and prediction.

        Args:
            y: true labels
            y_hat: output prediction

        Returns:
            mse
        """
        y_hat = y_hat
        error = (y - y_hat) ** 2
        return tf.sqrt(tf.reduce_mean(error, axis=axis))


### LSTM PREDICTOR ###
class RNNPredictor(tf.keras.Model):
    """ Predictor class to predict the specificity of peptide substrates."""

    def __init__(self, vocab_size, embedding_dim, rnn_units,
                 p, regu, max_len, output_len, mask_zero=False, training=True):
        super(RNNPredictor, self).__init__()
        self.training = training

        self.embedding = tf.keras.layers.Embedding(
            vocab_size, embedding_dim, input_length=max_len, mask_zero=mask_zero) #batch_input_shape=[batch_size, max_len])

        # LSTM
        # BLOCK 1: LSTM, DROP
        self.lstm = tf.keras.layers.LSTM(
            rnn_units, recurrent_regularizer=L2(regu), recurrent_dropout=p, return_sequences=True, stateful=False)
        self.lstm = tf.keras.layers.Bidirectional(self.lstm)
        self.drop = tf.keras.layers.Dropout(p)

        # BLOCK 2: LSTM, DROP
        self.lstm2 = tf.keras.layers.LSTM(
          rnn_units, recurrent_regularizer=L2(regu), recurrent_dropout=p, return_sequences=False, stateful=False)
        self.lstm2 = tf.keras.layers.Bidirectional(self.lstm2)
        self.drop2 = tf.keras.layers.Dropout(p)

        self.dense = tf.keras.layers.Dense(2 * rnn_units, activation='relu')
        self.dense_out = tf.keras.layers.Dense(output_len)

    def call(self, x):
        """ Forward pass through the model.

        Args:
            x: input to the model
        """
        x = self.embedding(x)
        x = self.lstm(x, training=self.training)
        if self.training:
            x = self.drop(x)
        x = self.lstm2(x, training=self.training)
        self.last_layer_embeddings = x
        if self.training:
            x = self.drop2(x)
        x = self.dense(x)
        x = self.dense_out(x)
        return x

    def compute_loss(self, y, y_hat):
        """ Computes loss, as mean absolute error, between true y and prediction.

        Args:
            y: true labels
            y_hat: output prediction

        Returns:
            absolute error
        """
        y_hat = y_hat #[:, -1, :] # batch, last, 1
        y_hat = tf.cast(y_hat, tf.float64)
        error = tf.abs(y - y_hat)
        return tf.reduce_mean(error)

    def compute_rmse(self, y, y_hat, axis=None):
        """Computes the root mean squared error between true y and prediction.

        Args:
            y: true labels
            y_hat: output prediction

        Returns:
            mse
        """
        y_hat = y_hat  # [:, -1, :]
        error = (y - y_hat) ** 2
        return tf.sqrt(tf.reduce_mean(error, axis=axis))


### LSTM GENERATOR MODELS ###
class RNNGenerator(tf.keras.Model):
    """ Generator class to learn how to predict new peptide substrates."""

    def __init__(self, batch_size, vocab_size, embedding_dim, rnn_units, p, regu, seq_len, training=True, num_layers=2):
        super(RNNGenerator, self).__init__()
        self.training = training
        self.vocab_size = vocab_size
        self.embedding_dim = embedding_dim
        self.num_layers = num_layers
        # Embed Inputs
        self.embedding = tf.keras.layers.Embedding(vocab_size, self.embedding_dim, batch_input_shape=[batch_size, seq_len])
        # NN
        self.lstm = [tf.keras.layers.LSTM(rnn_units, recurrent_regularizer=L2(regu), recurrent_dropout=p, return_sequences=True, stateful=True) for i in range(num_layers)]
        self.lstm = [tf.keras.layers.Bidirectional(self.lstm[i]) for i in range(num_layers)]
        self.drop = [tf.keras.layers.Dropout(p) for i in range(num_layers)]
        self.dense = tf.keras.layers.Dense(2 * rnn_units, activation='relu')
        self.dense_out = tf.keras.layers.Dense(vocab_size)

    def call(self, x):
        """ Forward pass through the model.

        Args:
            x: input to the model
        """
        x = self.embedding(x)
        x *= tf.math.sqrt(tf.cast(self.embedding_dim, tf.float32)) # set scale of layers
        for i in range(self.num_layers):
            x = self.lstm[i](x, training=self.training)
            if self.training:
                x = self.drop[i](x)
        self.last_layer_embeddings = x
        x = self.dense(x)
        x = self.dense_out(x)
        return x

    def compute_masked_loss(self, y, y_hat, mask):
        """ Computes loss between true y and prediction over masked locations"""
        # Select masked locations
        y_masked = tf.boolean_mask(y, mask)
        y_hat_masked = tf.boolean_mask(y_hat, mask)
        loss = tf.keras.losses.sparse_categorical_crossentropy(y_masked, y_hat_masked, from_logits=True)
        return tf.reduce_mean(loss)

    def compute_masked_accuracy(self, y, y_hat, mask):
        # Sample from predicted distribution to get characters
        m = tf.keras.metrics.CategoricalAccuracy()

        # Compute accuracy over masked tokens
        y_onehot = tf.one_hot(y, depth=self.vocab_size)
        y_masked = tf.boolean_mask(y_onehot, mask)
        y_hat_masked = tf.boolean_mask(y_hat, mask)
        # Compute accuracy
        m.update_state(y_masked, y_hat_masked)
        accuracy = m.result()
        return accuracy


class AutoregressiveRNN(tf.keras.Model):
    """ Generator class to generate new peptide substrates."""

    def __init__(self, batch_size, vocab_size, embedding_dim, rnn_units, p, regu, seq_len, training=True, mask_zero=False, num_layers=4):
        super(AutoregressiveRNN, self).__init__()
        self.training = training
        self.vocab_size = vocab_size
        self.embedding_dim = embedding_dim
        self.num_layers = num_layers
        # Embed Inputs
        self.embedding = tf.keras.layers.Embedding(vocab_size, self.embedding_dim,
                                                   batch_input_shape=[batch_size, seq_len], mask_zero=mask_zero)
        # NN
        self.lstm = [tf.keras.layers.LSTM(rnn_units, recurrent_regularizer=L2(regu), recurrent_dropout=p,
                                         return_sequences=True, stateful=True) for i in range(num_layers)]
        self.drop = [tf.keras.layers.Dropout(p) for i in range(num_layers)]
        # Output layer: Dense (fully-connected) layer that transforms the LSTM output to vocab size
        self.dense = tf.keras.layers.Dense(2 * rnn_units, activation='relu')
        self.dense_out = tf.keras.layers.Dense(vocab_size)

    def call(self, inputs):
        """ Forward pass through the model
        """
        x = self.embedding(inputs)
        x *= tf.math.sqrt(tf.cast(self.embedding_dim, tf.float32))  # set scale of layers
        for i in range(self.num_layers):
            x = self.lstm[i](x, training=self.training)
            if self.training:
                x = self.drop[i](x)
        self.last_layer_embeddings = x
        x = self.dense(x)
        x = self.dense_out(x)
        return x

    def compute_loss(self, y, y_hat):
        loss = tf.keras.losses.sparse_categorical_crossentropy(y, y_hat, from_logits=True)
        return tf.reduce_mean(loss)

    def compute_accuracy(self, y, y_hat):
        m = tf.keras.metrics.SparseCategoricalAccuracy()
        m.update_state(y, y_hat)
        return m.result()

def load_generator_model(model_type, training_scheme='unconditional', parent_dir=''):
    # Load relevant checkpoints, and model parameters
    batch_size = 1                                                                                                                                  

    if model_type == 'transformer':
        vocab_size = 22
        num_layers = 2
        num_heads = 6
        dropout = 0
        d_model = 64
        if training_scheme == 'unconditional':
            checkpoint_path = os.path.join(parent_dir+"weights/AUTOREG_transformer/unconditional/", "model.h5")
        elif training_scheme == 'conditional':
            num_layers=3
            checkpoint_path = os.path.join(parent_dir+"weights/AUTOREG_transformer/conditional/", "model.h5")
        elif training_scheme == 'both':
            num_layers=3
            checkpoint_path = os.path.join(parent_dir+"weights/AUTOREG_transformer/both/", "model.h5")
        elif training_scheme == 'rounded':
            num_layers=3
            d_model = 64
            checkpoint_path = os.path.join("weights/AUTOREG_transformer/both_rounded/", "model.h5") # rounded to tenth
        if training_scheme == 'unconditional':
            model = TransformerDecoder(
                num_layers=num_layers,
                d_model=d_model,
                num_heads=num_heads,
                dff=d_model,  # dense params
                vocab_size=vocab_size,
                dropout_rate=dropout)
        else:
            model = cleavenet.models.ConditionalTransformerDecoder(
                                num_layers=num_layers,
                                d_model=d_model,
                                num_heads=num_heads,
                                dff=d_model, # dense params
                                vocab_size=vocab_size,
                                dropout_rate=dropout)

    elif model_type == 'lstm':
        regu = 0.01
        d_model = 64
        dropout = 0.2
        d_embed = 64
        seq_len = 11
        num_layers = 4
        vocab_size=22
        model = AutoregressiveRNN(batch_size, vocab_size, d_embed, d_model, dropout,
                                                   regu, seq_len, training=False, mask_zero=False, num_layers=num_layers)
        checkpoint_path = os.path.join("weights/AUTOREG_lstm/20231016-133250_GEN/", "model.h5")

    if training_scheme == 'unconditional':
        model.summary()
        model.load_weights(checkpoint_path)
    return model, checkpoint_path

def load_predictor_model(model_type, checkpoint_path, batch_size, mask_zero=False):
    d_model = 32
    len_mmps = 18

    if model_type == 'lstm':
        embedding_dim = 22
        dropout = 0.25
        regu = 0.01
        max_len = 10
        vocab_size = 21  # did not train with CLS token
        model = cleavenet.models.RNNPredictor(vocab_size, embedding_dim, d_model,
                                              dropout, regu, max_len, len_mmps, mask_zero=mask_zero)

    elif model_type == 'transformer':
        d_model = 32
        num_layers = 2
        num_heads = 6
        dropout = 0
        vocab_size = 22
        embedding_dim = 32
        model = cleavenet.models.TransformerEncoder(
            num_layers=num_layers,
            d_model=embedding_dim, #d_model,
            num_heads=num_heads,
            dff=d_model,  # dense params
            vocab_size=vocab_size,
            dropout_rate=dropout,
            output_dim=len_mmps,
            pool_outputs=True,
            mask_zero=mask_zero)

    fake_batch = np.array([[21, 20, 14, 8, 9, 13, 10 , 9 ,16, 17 , 3]])
    model(fake_batch, training=False) # build model in TF
    model.summary()
    model.load_weights(checkpoint_path)  # load weights
    return model


def inference(model, dataloader, causal=False, seq_len=10, penalty=1, verbose=False, conditioning_tag=None, temperature=1):
    if causal: # autoregressive inference
        if conditioning_tag is None:
            start_seq = np.array([dataloader.char2idx[dataloader.START]], dtype=np.int32)  # start from START token
            generated_seq = tf.expand_dims(start_seq, 0)
        else: 
            tag = np.array(conditioning_tag) # start from Z-scores
            generated_seq = (np.array([[]]), tag) # seq, tags
        reach_stop = False
        for i in range(seq_len):
            if reach_stop == False:
                prediction = model(generated_seq, training=False)
                predicted_id = tf.random.categorical(prediction[:, -1, :]/temperature, num_samples=1, dtype=tf.int32)
                if penalty > 1:
                    if conditioning_tag is None:
                        if predicted_id == generated_seq[0][i]:
                            penalized_prediction = prediction.numpy()
                            penalized_prediction[:, -1, int(predicted_id)] = prediction[0, -1, int(predicted_id)]/penalty
                            predicted_id = tf.random.categorical(penalized_prediction[:, -1, :], num_samples=1, dtype=tf.int32)
                    else:
                        if generated_seq[0].shape[1] > 0:
                            if predicted_id == generated_seq[0][0][i-1]:
                                penalized_prediction = prediction.numpy()
                                penalized_prediction[:, -1, int(predicted_id)] = prediction[0, -1, int(predicted_id)]/penalty
                                predicted_id = tf.random.categorical(penalized_prediction[:, -1, :], num_samples=1, dtype=tf.int32)
                    
                if conditioning_tag is None:
                    generated_seq = tf.concat([generated_seq, predicted_id], 1)
                else:
                    generated_seq = (np.expand_dims(np.append(generated_seq[0], predicted_id), 0), tag)   
                
                if verbose:
                    if conditioning_tag is None:    
                        print("step", i, "seq", cleavenet.data.untokenize_sequences(generated_seq, dataloader))
                    else:
                        print("step", i, "seq", cleavenet.data.untokenize_sequences(generated_seq[0].astype(int), dataloader))
                
                if predicted_id == dataloader.char2idx[dataloader.STOP]:
                    reach_stop = True
            else:
                break
        if conditioning_tag is None:
            generated_seq = generated_seq[0]
        else:
            generated_seq = generated_seq[0][0][:-1].astype(int)
    else: # MLM inference
        generated_seq = np.zeros((seq_len), dtype=int) + dataloader.char2idx[dataloader.MASK] # start from all mask token
        loc = np.arange(seq_len)
        np.random.shuffle(loc)  # shuffle order
        for i in loc:
            predictions = model(tf.expand_dims(generated_seq, 0), training=False)
            predicted_id = tf.random.categorical(predictions[:, i, :], num_samples=1)
            generated_seq[i] = predicted_id
            print("step",i, "out seq", cleavenet.data.untokenize_sequences(generated_seq, dataloader))

    return generated_seq


def prediction(data_path_kukreja, gen_data, generated_dir, true_zscores=None, true_mmps=None, checkpoint_dir='weights/', predictor_model_type='transformer', number_top_candidates=50):
    if not os.path.exists(generated_dir):
        os.mkdir(generated_dir)
    if predictor_model_type == 'transformer':
        ensembles = ['transformer_0/',
                 'transformer_1/',
                 'transformer_2/',
                 'transformer_3/',
                 'transformer_4/'
                 ]
    elif predictor_model_type == 'lstm':
        print("USING LSTM PREDICTOR FOR EVALUATION")
        ensembles = [
            'lstm_0/',
            'lstm_1/',
            'lstm_2/',
            'lstm_3/',
            'lstm_4/'
        ]
    
    kukreja = cleavenet.data.DataLoader(data_path_kukreja, seed=0, task='regression', model=predictor_model_type, test_split=0.2,
                                        dataset='kukreja')
    
    max_seq_len = max([len(s) for s in gen_data])
    gen_data = [cleavenet.data.pad(seq, max_seq_len=max_seq_len, pad_token='-') for seq in gen_data]
    x_all = cleavenet.data.tokenize_sequences(gen_data, kukreja)
    if predictor_model_type == 'transformer':
        cls_idx = kukreja.char2idx[kukreja.CLS]
        x_all = np.stack([np.append(np.array(cls_idx), s) for s in x_all])
    num_samples = len(gen_data)
    batch_size = num_samples
    print(num_samples, gen_data[0])
    predictions = []
    for e_num, ensemble in enumerate(ensembles):
        print("Running", e_num, ensemble)
        print("EVALUATING SEQUENCES FROM", generated_dir)
        checkpoint_path = os.path.join(checkpoint_dir, ensemble, "model.h5")
        # Build and load predictor model
        model = cleavenet.models.load_predictor_model(model_type=predictor_model_type, checkpoint_path=checkpoint_path,
                                                      batch_size=batch_size, mask_zero=True)
        # Predict z-scores
        y_hat = model(x_all, training=False)  # forward pass
        if true_zscores is not None:
            subset_preds = []
            for i, m in enumerate(mmps):
                if m in true_mmps:
                    j = true_mmps.index(m)
                    subset_preds.append(y_hat[:, i])
            subset_preds = np.stack(subset_preds, axis=1)
            mae = tf.reduce_mean(tf.abs(true_zscores - subset_preds), axis=0)
            plotter.plot_mae(mae, true_mmps, generated_dir + str(e_num) + '_')
            rmse = model.compute_rmse(true_zscores, subset_preds, axis=0)
            plotter.plot_rmse(rmse, true_mmps, generated_dir + str(e_num) + '_')
            plotter.plot_parity(true_zscores, subset_preds, true_mmps, generated_dir)
        predictions.append(y_hat)
        if e_num == (len(ensembles)-1): # save embeddings from last ensemble model for plotting later
            embeddings = model.last_layer_embeddings
            np.save(os.path.join(generated_dir, 'embeddings.npy'), np.array(embeddings))

    predictions = np.stack(np.array(predictions))
    print(predictions.shape)
    means, std = analysis.confidence_score(predictions, mmps)

    if true_zscores is not None:
        print("plot true vs mean predicted")
        pred = []
        for i, m in enumerate(mmps):
            if m in true_mmps:
                print(m)
                j = true_mmps.index(m)
                pred.append(y_hat[:, i])
                scores = analysis.save_to_dataframe(x_all, true_zscores[:, j], i, means, std, z_cutoff=0,
                                                    write_top_scores=True,
                                                    dataloader=kukreja, mmp=m, save_path=generated_dir)
                plotter.true_pred_ranked_scatter_z(scores, generated_dir, m)
                plotter.confidence_ranked_scatter_z(scores, generated_dir, m)
                plotter.confusion(scores['Cleaved true'], scores['Cleaved pred'], generated_dir, m)
                plotter.plot_auc(true_zscores[:,j], y_hat[:,i], i, mmps, generated_dir)
        pred = np.stack(pred, axis=1)

    
    print("Calculated confidence")
    analysis.eval_all_mmp(means, x_all, kukreja, generated_dir, z_score_cutoff=0)
    
    for i,m in enumerate(mmps):
        if i == 0:
            if os.path.exists(os.path.join(generated_dir, 'weighted_all_scores.csv')):
                os.remove(os.path.join(generated_dir, 'weighted_all_scores.csv'))
            if os.path.exists(os.path.join(generated_dir, 'all_uncertainty.csv')):   
                os.remove(os.path.join(generated_dir, 'all_uncertainty.csv'))
            if os.path.exists(os.path.join(generated_dir, 'all_scores.csv')):   
                os.remove(os.path.join(generated_dir, 'all_scores.csv'))
        print("Iterating over", m)
        scores = analysis.save_to_dataframe(x_all, None, i, means, std, z_cutoff=0, write_top_scores=True,
                                            find_matches=True, dataloader=kukreja, mmp=m, save_path=generated_dir,
                                            top=number_top_candidates)
        plotter.confidence_ranked_scatter_z(scores, generated_dir, m)
        confidence_threshold = plotter.confidence_histogram(scores, generated_dir, m)
        plotter.confidence_ranked_scatter_z(scores, generated_dir, m, threshold=confidence_threshold)

    return means, std


def predict_scores_simple(substrates, checkpoint_dir='../weights/', save_dir='outputs/', model_architecture='transformer'):
    data_dir = cleavenet.utils.get_data_dir()
    data_path = os.path.join(data_dir, "kukreja.csv")
    pred_zscores, std_zscores = cleavenet.models.prediction(data_path,
                                                            substrates,
                                                            save_dir,
                                                            predictor_model_type=model_architecture,
                                                            checkpoint_dir=checkpoint_dir)
    return pred_zscores, std_zscores

def simple_inference(num_seqs, repeat_penalty, temperature):
    data_dir = cleavenet.utils.get_data_dir()
    data_path = os.path.join(data_dir, "kukreja.csv")
    kukreja = cleavenet.data.DataLoader(data_path, seed=0, task='generator', model='autoreg', test_split=0.2,
                                                dataset='kukreja')
    # From dataloader get necessary variables
    start_id = kukreja.char2idx[kukreja.START]
    # Load model
    model, checkpoint_path = cleavenet.models.load_generator_model(model_type='transformer',
                                                                   training_scheme='rounded')
    # Fake run to load data and build model
    conditioning_tag_fake = [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]
    generated_seq = cleavenet.models.inference(model, kukreja, causal=True, seq_len=11,
                                               penalty=1, # no penalty
                                               verbose=False,
                                               conditioning_tag=conditioning_tag_fake,
                                               temperature=1 # no temp
                                               )
    conditioning_tag = [[start_id]] # unconditional generation
    untokenized_seqs = []
    for i in range(len(conditioning_tag)):
        for j in range(num_seqs):
            model.built=True
            model.load_weights(checkpoint_path)  # Load model weights
            # Generate using loaded weights
            generated_seq = cleavenet.models.inference(model, kukreja, causal=True, seq_len=11,
                                                       penalty=repeat_penalty,
                                                       verbose=False,
                                                       conditioning_tag=[conditioning_tag[i]], temperature=temperature)
            #tokenized_seqs.append(generated_seq)
            untokenized_seqs.append(''.join(kukreja.idx2char[generated_seq]))
    return untokenized_seqs


import os
import csv

originial_mmps = ['MMP1', 'MMP2', 'MMP3', 'MMP7', 'MMP8', 'MMP9', 'MMP10', 'MMP11',
        'MMP12', 'MMP13', 'MMP14', 'MMP15', 'MMP16', 'MMP17', 'MMP19', 'MMP20',
        'MMP24', 'MMP25']

mmps = ['MMP1', 'MMP10', 'MMP11', 'MMP12', 'MMP13', 'MMP14', 'MMP15', 'MMP16',
       'MMP17', 'MMP19', 'MMP2', 'MMP20', 'MMP24', 'MMP25', 'MMP3', 'MMP7',
       'MMP8', 'MMP9'] # the pivot table re-orderes MMPs in the creation of the train/test splits

bhatia_mmps = ['MMP1', 'MMP10', 'MMP12', 'MMP13', 'MMP17', 'MMP3', 'MMP7']

def get_data_dir():
    this_dir = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(this_dir, "../data")

def get_save_dir():
    this_dir = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(this_dir, "save")

def save_sequences(substrates, save_dir='outputs/'):
    save_file = os.path.join(save_dir, 'sequences.csv')
    os.makedirs(save_dir, exist_ok=True)
    with open(save_file, 'w', newline='') as f:
        writer = csv.writer(f)
        writer.writerows(substrates)

def import_tensorflow():
    # Disable this function in __init__ if you dont know what you are doing!!
    # Filter tensorflow version warnings -
    import os
    # https://stackoverflow.com/questions/40426502/is-there-a-way-to-suppress-the-messages-tensorflow-prints/40426709
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # or any {'0', '1', '2'}
    import warnings
    # https://stackoverflow.com/questions/15777951/how-to-suppress-pandas-future-warning
    warnings.simplefilter(action='ignore', category=FutureWarning)
    warnings.simplefilter(action='ignore', category=Warning)
    import tensorflow as tf
    tf.get_logger().setLevel('INFO')
    tf.autograph.set_verbosity(0)
    import logging
    tf.get_logger().setLevel(logging.ERROR)
    return tf


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

def parse_requirements(filename):
    with open(filename, 'r') as file:
        return file.read().splitlines()

setuptools.setup(
    name="cleavenet", # Replace with your own username
    version="0.1.2",
    description="Deep learning model for substrate cleavage prediction and generation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/cleavenet",
    license='BSD 2-clause',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=parse_requirements('requirements.txt')
)



import os

from collections import Counter
import jellyfish
import numpy as np
import pandas as pd
import scipy.stats as st

import cleavenet
from cleavenet.utils import mmps

def confidence_score(predictions, mmps):
    "calculate mean, standard error for each mmp"
    means = np.zeros((predictions.shape[1], len(mmps)))
    st_devs = np.zeros((predictions.shape[1], len(mmps)))

    for i, m in enumerate(mmps):
        means[:,i] = np.mean(predictions[:, :, i], axis=0)  # Take mean and std across batches
        st_devs[:,i] = np.std(predictions[:, :, i], axis=0)

    return means, st_devs


def confidence_interval_1_sided(std, num_samples, confidence_level=0.80):
    z_score = st.norm.ppf(confidence_level)
    ci = z_score * (std/np.sqrt(num_samples))
    return ci


def save_to_dataframe(x_all, y_all, mmp_idx, y_hat_all, uncertainty, z_cutoff=0, write_top_scores=False,
                      find_matches=False, dataloader=None, mmp=None, save_path=None, threshold=0, top=50):
    labels = pd.DataFrame(data=x_all, index=np.arange(len(x_all)))
    if y_all is not None:
        labels['True Z-scores'] = y_all
    labels['Pred Z-scores'] = y_hat_all[:,mmp_idx]
    labels['uncertainty'] = uncertainty[:,mmp_idx]
    # make df with true and predicted labels only
    if y_all is not None:
        scores = labels.iloc[:, -3:]
    else:
        scores = labels.iloc[:, -2:]
    if y_all is not None:
        scores['Cleaved true'] = scores['True Z-scores'].apply(lambda x: 1 if x > z_cutoff else z_cutoff)
        scores['Cleaved pred'] = scores['Pred Z-scores'].apply(lambda x: 1 if x > z_cutoff else z_cutoff)
    if write_top_scores:
        # Efficient (raw top 100 z-scores)
        sequences = ["".join(dataloader.idx2char[s]) for s in x_all]
        scores['sequences'] = sequences
        scores = scores.set_index('sequences')
        if threshold > 0:
            scores_thresh = scores[scores['uncertainty'] <= threshold]
            save_file1 = '_top'+str(top)+'_cleaved_threshold.csv'
            save_file2 = '_weighted_top'+str(top)+'_cleaved_threshold.csv'
            scores_sorted = scores_thresh.sort_values(by=['Pred Z-scores'], ascending=False)[:top]
        else:
            save_file1 = '_top'+str(top)+'_cleaved.csv'
            save_file2 = '_weighted_top'+str(top)+'_cleaved.csv'
            scores_sorted = scores.sort_values(by=['Pred Z-scores'], ascending=False)[:top]
        if find_matches:
            exact_match_freq, active_match_freq = exact_match(list(scores_sorted.index), dataloader)
            # save all to dataframe
            scores_sorted['exact matches'] = exact_match_freq
            scores_sorted['exact active'] = active_match_freq
        scores_sorted['mmp'] = [mmp] * len(scores_sorted)
        scores_sorted.to_csv(os.path.join(save_path, mmp+save_file1))

        # Selective (relative to mean top z-score)
        all_mmp_idx = list(range(y_hat_all.shape[1]))
        y_hat_weighted = np.zeros((y_hat_all.shape))
        # Weight scores based on mean of all other MMP scores (z-score the z-scores)
        for idx in all_mmp_idx:
            idx_to_mean = [i for i in all_mmp_idx if i is not idx]
            y_hat_weighted[:,idx] = y_hat_all[:,idx] - np.mean(y_hat_all[:, idx_to_mean], axis=1)
        scores_weighted = pd.DataFrame(y_hat_weighted[:,mmp_idx], columns=['Weighted Z-scores'])
        scores_weighted['uncertainty'] = uncertainty[:,mmp_idx]
        scores_weighted['sequences'] = sequences
        scores_weighted = scores_weighted.set_index('sequences')
        if threshold > 0:
            scores_weighted_thresh = scores_weighted[scores_weighted['uncertainty'] <= threshold]
            weighted_scores_sorted = scores_weighted_thresh.sort_values(by=['Weighted Z-scores'], ascending=False)[:top]
        else:
            weighted_scores_sorted = scores_weighted.sort_values(by=['Weighted Z-scores'], ascending=False)[:top]
        if find_matches:
            exact_match_freq, active_match_freq  = exact_match(list(weighted_scores_sorted.index), dataloader)
            # save all to dataframe
            weighted_scores_sorted['exact matches'] = exact_match_freq
            weighted_scores_sorted['exact active'] = active_match_freq
        weighted_scores_sorted['mmp'] = [mmp] * len(weighted_scores_sorted)
        weighted_scores_sorted.to_csv(os.path.join(save_path, mmp + save_file2))

        # Write all scores once
        if not os.path.exists(os.path.join(save_path, 'weighted_all_scores.csv')):
            print("writing all scores")
            all_scores = pd.DataFrame(y_hat_weighted, columns=mmps)
            all_scores['sequences'] = sequences
            all_scores = all_scores.set_index('sequences')
            all_scores.to_csv(os.path.join(save_path, 'weighted_all_scores.csv'))

        if not os.path.exists(os.path.join(save_path, 'all_scores.csv')):
            print("writing all scores")
            all_scores = pd.DataFrame(y_hat_all, columns=mmps)
            all_scores['sequences'] = sequences
            all_scores = all_scores.set_index('sequences')
            all_uncertainty = pd.DataFrame(uncertainty, columns=mmps)
            all_uncertainty['sequences'] = sequences
            all_uncertainty = all_uncertainty.set_index('sequences')
            all_scores.to_csv(os.path.join(save_path, 'all_scores.csv'))
            all_uncertainty.to_csv(os.path.join(save_path, 'all_uncertainty.csv'))
    return scores


def exact_match(sequences, dataloader):
    """
    sequences (str) : list of sequences to compare
    train (str) : list of train data sequences
    ensemble: pick 1 run
    """
    x_train = dataloader.X # use all not just train here

    dist_array = np.zeros((len(sequences), len(x_train)))
    active_array = np.zeros((len(sequences), len(x_train)))
    #s_array = np.zeros((len(sequences)))

    exact_match_freq = []
    active_match_freq = []

    for i in range(len(sequences)):
        full_matches = 0
        active_matches = 0
        #s_array[i] = jellyfish.levenshtein_distance(sequences[i], s1)
        for j in range(len(x_train)):
            active_array[i, j] = jellyfish.levenshtein_distance(sequences[i][2:7], x_train[j][2:7])
            dist_array[i, j] = jellyfish.levenshtein_distance(sequences[i], x_train[j])
        if 0 in dist_array[i, :]:
            index = np.where(dist_array[i] == 0)
            full_matches+=len(index)
        if 0 in active_array[i, :]:
            index_a = np.where(active_array[i] == 0)
            active_matches+=len(index_a)
        exact_match_freq.append(full_matches)
        active_match_freq.append(active_matches)
    return exact_match_freq, active_match_freq #, s_array


def eval_all_mmp(means, x_all, dataloader, save_path, z_score_cutoff=0):
    sequences = ["".join(dataloader.idx2char[s]) for s in x_all]
    df = pd.DataFrame(means, columns=mmps)
    df['freq'] = [0] * len(df)
    for mmp in mmps:
        freq_index = df[df[mmp] >= z_score_cutoff].index
        df['freq'].iloc[freq_index] += 1
    df['sequences'] = sequences
    df = df.set_index('sequences')
    df = df.sort_values('freq', ascending=False)[:20]
    df.to_csv(os.path.join(save_path, 'allmmp_top20_cleaved.csv'))


def extract_sequences(x_train, dataloader):
    seqs_arr = []
    x_train = cleavenet.data.tokenize_sequences(x_train, dataloader)
    for k in range(x_train.shape[1]):
        seqs = ''
        for s in x_train[:, k]:
            tokenized = dataloader.idx2char[s]
            for t in tokenized:
                if t != '-':
                    seqs += t
        seqs_arr.append(seqs)
    return seqs_arr


def count_by_res(seqs_arr):
    dict_arr = []
    for i in seqs_arr:
        aminos = Counter(
             {'A': 0, 'C': 0, 'D': 0, 'E': 0, 'F': 0, 'G': 0, 'H': 0, 'I': 0, 'K': 0, 'L': 0, 'M': 0, 'N': 0, 'P': 0,
              'Q': 0, 'R': 0, 'S': 0, 'T': 0, 'V': 0, 'W': 0, 'Y': 0})
        aminos.update(i)
        dict_arr.append(aminos)
    return dict_arr


def normalize(list):
    norm = sum(list)
    new_list = [item / norm for item in list]
    return new_list


def calc_selectivity_score(mmps, df, penalize=[]):
    df_new = pd.DataFrame()
    for mmp in mmps:
        current_mmp = [mmp]
        other_mmps = [m for m in mmps if m not in current_mmp and m not in penalize]
        ss = np.array(df[current_mmp]).squeeze()-(df[other_mmps].mean(axis=1))
        if len(penalize) > 0:
            ss = ss - np.array(df[penalize].mean(axis=1).squeeze())
        df_new[str(mmp)] = ss
    return df_new


def get_roc_data(y, y_hat, true_z_cutoff=2, width=0.1):
    x_values = []
    y_values = []
    df = pd.DataFrame()
    df['True'] = y #[:, mmp_index]
    df['Pred'] = y_hat #[:, mmp_index]
    thresh_range = np.arange(y.min(), y.max(), width)
    for z_cutoff in thresh_range:
        # Make scores binary according to 2 z-thresh
        df['Cleaved True'] = df['True'].apply(lambda x: 1 if x > true_z_cutoff else 0)
        df['Cleaved Pred'] = df['Pred'].apply(lambda x: 1 if x > z_cutoff else 0)
        # Conditions for TP, FP, FN, TN
        def conditions(s):
            if (s['Cleaved True'] == 1) & (s['Cleaved Pred'] == 1):
                return 'TP'
            elif ((s['Cleaved True'] == 0) & (s['Cleaved Pred'] == 1)):
                return 'FP'
            elif (s['Cleaved True'] == 1) & (s['Cleaved Pred'] == 0):
                return 'FN'
            else:
                return 'TN'
        df['class'] = df.apply(conditions, axis=1)
        #(df.head())
        tp = (df['class'] == 'TP').sum()
        fn = (df['class'] == 'FN').sum()
        tn = (df['class'] == 'TN').sum()
        fp = (df['class'] == 'FP').sum()
        tpr = tp/(tp+fn)
        fpr = fp/(tn+fp)
        y_values.append(tpr)
        x_values.append(fpr)
    return x_values, y_values


def calc_auc(x_values, y_values):
    inds = [x for i, x in enumerate(zip(x_values, y_values)) if x[0] != x[1]]
    area = 0
    for i in range(len(inds)-1):
        area += (inds[i][0] - inds[i+1][0]) * (inds[i][1] + inds[i+1][1]) / 2 # calc using trapezoid
    return area


import os
import random

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from cleavenet.utils import mmps

def pad(sequence, max_seq_len=10, pad_token='-'):
    seq_len = len(sequence)
    if seq_len <= max_seq_len:
        sequence += pad_token * (max_seq_len-seq_len)
    elif seq_len > max_seq_len:
        print("Increase max seq len for padding")
        import pdb; pdb.set_trace()
    return sequence


def custom_round(x, base=5):
    return base * round(float(x)/base)


class DataLoader(object):
    """
    DataLoader class for loading peptide sequence data, given a CSV file (data_path)
    Creates 1 test/train split for Dataset, or loads saved splits
    Saves a char2idx dict mapping for each model
    """
    def __init__(self, data_path, seed=0, task='predictor', model='bert', test_split=0.2, dataset='kukreja',
                 use_dataloader=None, rounded=False):
        self.seed = seed
        self.model = model  # bert, autoregressive, regression
        self.dataset = dataset
        if rounded:
            self.dataset += '_rounded'
        self.data_path = data_path
        self.test_split = test_split
        self.task = task
        np.random.seed(self.seed)
        random.seed(self.seed)

        # Set save paths
        self.out_path = os.path.join('splits/', self.dataset+'/') # One split per dataset
        if os.path.exists(self.out_path):
            print("Splits previously written to file")
            self.X = list(get_data(self.out_path + 'X_all.csv', names=['sequence']).index)
            self.y = get_data(self.out_path + 'y_all.csv', index_col=None, names=mmps).values
            self.sequences = self.X
            if test_split > 0:
                self.X_train = list(get_data(self.out_path + 'X_train.csv', names=['sequence']).index)
                self.y_train = get_data(self.out_path + 'y_train.csv', index_col=None, names=mmps).values
                self.X_test = list(get_data(self.out_path + 'X_test.csv', names=['sequence']).index)
                self.y_test = get_data(self.out_path + 'y_test.csv', index_col=None, names=mmps).values
        else:
            os.makedirs(self.out_path)
            print('Splits directory created', self.out_path)
            #Load data
            data = self.load_zscore_data()
            self.sequences = data.index.to_list()
            if dataset == 'kukreja':
                # replace gaps in data in kukreja
                # these are artificially created in csv processing
                self.sequences = [seq.replace(' ', '') for seq in self.sequences]
            if rounded:
                data = data.apply(lambda x: custom_round(x, base=0.5)) # round to nearest 0.5
            # Assign data
            self.X = self.sequences  # sequences
            self.y = data.values
            np.savetxt(self.out_path + 'X_all.csv', self.X, delimiter=",", fmt='%s')
            np.savetxt(self.out_path + 'y_all.csv', self.y, delimiter=",")
            if test_split > 0:
                self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(self.X, self.y,
                                                                                        test_size=self.test_split,
                                                                                        random_state=self.seed)
                np.savetxt(self.out_path + 'X_train.csv', self.X_train, delimiter=",", fmt='%s')
                np.savetxt(self.out_path + 'y_train.csv', self.y_train, delimiter=",")
                np.savetxt(self.out_path + 'X_test.csv', self.X_test, delimiter=",", fmt='%s')
                np.savetxt(self.out_path + 'y_test.csv', self.y_test, delimiter=",")

        # Create vocab
        if not use_dataloader:
            self.char2idx, self.idx2char = self.create_vocab()
            print("Vocab: \n", self.char2idx)
        else:
            self.char2idx = use_dataloader.char2idx
            self.idx2char = use_dataloader.idx2char
            print("Vocab: \n", self.char2idx)

        # If using generator
        if task == 'generator':
            train_data = pd.DataFrame(self.y_train, columns=mmps)
            train_data['sequence'] = self.X_train
            train_data = train_data.set_index('sequence')
            self.X_train = train_data.index.to_list()
            self.y_train = train_data.values

            test_data = pd.DataFrame(self.y_test, columns=mmps)
            test_data['sequence'] = self.X_test
            test_data = test_data.set_index('sequence')
            self.X_test = test_data.index.to_list()
            self.y_test = test_data.values


    def load_zscore_data(self):
        """
        Load z-scored protease-substrate data from path
        """
        df_z = pd.read_csv(self.data_path, header=0)
        df_z = df_z.dropna(how='all', axis=1)
        if self.dataset == 'kukreja' or self.dataset == 'kukreja_rounded':
            df_z = df_z.drop('construct', axis=1)
        elif self.dataset == 'bhatia': # Remove all seqs not == 10 residues in length (this is outside training task)
            new_sequences = df_z['sequence'].to_list()
            new_sequences = [seq[2:-2] for seq in new_sequences] # crop in middle
            seq_lens = [len(seq) for seq in new_sequences]
            df_z['sequence'] = new_sequences
            df_z['len'] = seq_lens
            df_z = df_z.drop('len', axis=1)
            print(df_z)
        df_z = pd.pivot_table(df_z, index=["sequence"])  # average z-scores for any duplicated sequences
        print("WARNING: PIVOT TABLE WILL RE-ORDER COLUMNS IN DATAFRAME- ENSURE LABELING IS MAPPED CORRECTLY")
        return df_z


    def create_vocab(self):
        # Find unique characters
        seqs_joined = "".join(self.sequences)

        if self.task == 'regression':
            self.PAD = '-'
            vocab = [self.PAD] # set 0 to pad token
            vocab += sorted(set(seqs_joined))
            # Add CLS for transformer
            if self.model == 'transformer':
                self.CLS = '!' # classifier token
                vocab += self.CLS
        else:
            vocab = sorted(set(seqs_joined))
        # For generator task add special tokens
        if self.task == 'generator':
            if self.model == 'bert':
                self.MASK = '#'
                vocab += self.MASK
            elif self.model == 'autoreg':
                self.START = '$'
                self.STOP = '*'
                vocab += self.START
                vocab += self.STOP

        char2idx = {u: i for i, u in enumerate(vocab)}
        idx2char = np.array(vocab)
        return char2idx, idx2char


def tokenize_sequences(sequences, dataloader):
    """
    Tokenize sequences
    """
    char2idx, idx2char = dataloader.char2idx, dataloader.idx2char
    tokenized = [np.array([char2idx[aa] for aa in sequence]) for sequence in sequences]
    return np.stack(tokenized) # This will break if sequences are different lengths


def untokenize_sequences(sequences, dataloader):
    """
    Untokenize sequences
    """
    char2idx, idx2char = dataloader.char2idx, dataloader.idx2char
    untokenized = [np.array(idx2char[sequence]) for sequence in sequences]
    return np.stack(untokenized) # This will break if sequences are different lengths


def get_batch(x, y, batch_size, dataloader, test=False, transformer=False):
    """
    Randomly sample batch of training data (x,y), and tokenize inputs (x) for training
    """
    #x = np.array(x)
    if transformer:
        cls_idx = dataloader.char2idx[dataloader.CLS]

    if test==True:
        batch_X = x
        batch_Y = y
    else:
        # Randomly choose the indices for the examples in the training batch
        selected_inds = np.random.choice(len(x), size=batch_size)
        batch_X = [x[s] for s in selected_inds]
        batch_Y = [y[s] for s in selected_inds]

    # Tokenize inputs
    tokenized_batch_X = tokenize_sequences(batch_X, dataloader)
    if transformer:
        tokenized_batch_X = np.stack([np.append(np.array(cls_idx), s) for s in tokenized_batch_X])
    return tokenized_batch_X, batch_Y


def get_data(path, index_col=0, names=None):
    """Transform data from .csv file to np.array
    Input:
        path (str): path to .csv
    """
    if names is not None:
        data = pd.read_csv(path, index_col=index_col, names=names)
    else:
        data = pd.read_csv(path, index_col=index_col)
    return data


def read_csv(path):
    data = pd.read_csv(path, index_col=0, names=['sequences'])
    return list(data.index)


def get_masked_batch(seqs, batch_size, rng, dataloader):
    """ Corrupt inputs with a BERT-like training scheme

    Args:
        seqs (np.array): input sequences to be divided into kmers
        batch_size (int)
        rng: ranodom seed

        Returns:
        masked_seq (array): corrupted sequences
        target_seq (array): original sequences
        mask (array): bool array of corrupted locations
    """
    # Initiate arrays to store input and output sequences for all training sequences
    target_seq = []
    masked_seq = []
    mask_output = []

    vocab_words = list(dataloader.char2idx.keys())
    mask_id = dataloader.char2idx[dataloader.MASK]

    batch_inds = np.random.choice(len(seqs), size=batch_size)
    batch_seqs = [seqs[i] for i in batch_inds]
    batch_seqs = tokenize_sequences(batch_seqs, dataloader)

    for seq in batch_seqs:
        # calculate seq length
        n = len(seq)
        # For now randomly mask 15% of sequence with mask token only
        num_mask = 2 #int(0.15 * n) # ~ this is approx 2 for a 10mer
        # generate mask
        mask_arr = np.random.choice(n, num_mask, replace=False)  # Generates array of len num_mask
        index_arr = np.arange(0, n)  # index array [1...seq_len]
        mask = np.isin(index_arr, mask_arr, invert=False).reshape(index_arr.shape)  # mask bools indices specified by mask_arr
        # Mask inputs
        # 80% of time w/ mask token
        # 10% random token
        # 10% with same token
        masked_seq_temp = seq.copy()
        for masked_index in mask_arr:
            #print(masked_index)
            if rng.random() < 0.8:
                masked_token = mask_id
            else:
                # 10% of the time, keep original
                if rng.random() < 0.5:
                    masked_token = seq[masked_index]
                # 10% of the time, replace with random token
                else:
                    masked_token = rng.randint(0, len(vocab_words) - 1)
            masked_seq_temp[masked_index] = masked_token
        masked_seq.append(masked_seq_temp)
        mask_output.append(mask)
        target_seq.append(seq)
    return np.array(masked_seq), np.array(target_seq), np.array(mask_output) # input (x), target(y), mask


def get_autoreg_batch(seqs, batch_size, dataloader, width=11, conditioning_tag=None, rng=0, randomize_tag=False):
    """
    Autoregressive inputs/target
    seqs: input sequences (list of strings)
    batch_size: batch size (int)
    dataloader: saved dataloader with dictionary (Dataloader)
    width: context window (for short seqs, this is the entire length (10) + start or stop token (1)
    conditioning_tag: array of z scores in same order as sequences (y_train), or None
    rng: random seed 

    Returns:
        src_seqs (array): [START] + seqs[0:10]
        target_seq (array): seqs shifted by 1; seqs[1:10] + [END]
    """
    start_idx = dataloader.char2idx[dataloader.START]
    stop_idx = dataloader.char2idx[dataloader.STOP]

    batch_inds = np.random.choice(len(seqs), size=batch_size)
    batch_seqs = [seqs[i] for i in batch_inds]
    batch_seqs = tokenize_sequences(batch_seqs, dataloader)

    # Add special chars
    batch_seqs = [np.append(np.append(np.array(start_idx), s), np.array(stop_idx)) for s in batch_seqs]
    slice_index = np.random.randint(0, len(batch_seqs[0])-width, size=len(batch_seqs)) # if context window less than length (not needed here)
    
    target_seq = [s[slice_index[i]+1:slice_index[i]+width+1] for i, s in enumerate(batch_seqs)]
    if conditioning_tag is not None:
        batch_tags = [np.array(conditioning_tag[i]) for i in batch_inds]
        if randomize_tag:  
            batch_tags = [np.array(b) if rng.random() < 0.5 else np.array([start_idx]) for b in batch_tags]
        batch_seqs = [s[1:-1] for s in batch_seqs]
        return (np.array(batch_seqs), batch_tags), np.array(target_seq)
    else:
        # sample around window width
        src_seqs = [s[slice_index[i]:slice_index[i]+width] for i, s in enumerate(batch_seqs)]
        return np.array(src_seqs), np.array(target_seq)